/*
	VB ROM-image Shrinker (the opposite of a "padder")
	(c) 2010 Amos "RunnerPack" Bieler.  All rights reserved.  GPL licensed.
*/

#include <stdlib.h>
#include <stdio.h>

#define HEADER_SIZE 544

typedef unsigned long u32;

u32 ceilp2(u32 x) {
	u32 y = x - 1;
	y |= (y >> 1);
	y |= (y >> 2);
	y |= (y >> 4);
	y |= (y >> 8);
	y |= (y >> 16);
	y += 1;
	return y;
}

int main(int argc, char *argv[]) {

	FILE *fin, *fout;				/* File pointers */
	char t[13] = "temptemp.tmp";	/* Temporary filename (if output name isn't specified) */
	u32 size, new_size;				/* Current and desired sizes of the file. */
	u32 i = 0;						/* Miscellaneous iterator */
	int pad1, pad2;					/* Samples of the possible padding */

	/* If there aren't enough arguments, give the user the 411. */
	if (argc < 2) {
		printf("----------------------\nVB ROM-image Shrinker\n");
		printf("(c) 2003 Amos \"RunnerPack\" Bieler\n----------------------\n");
		printf("Usage: shrinkvb.exe <infile> [<outfile>]\n\n");
		printf("The size of <infile> is halved if it is at least half padding.\n\n");
		printf("It is then saved to <outfile> if specified, otherwise <infile> is overwritten.\n");
		printf("THERE'S NO GUARANTEE THIS WON'T BREAK THE ROM, SO MAKE A BACKUP!\n");
		printf("The padding is auto-detected, as long as it's the same byte repeated.\n\n");

		exit(1);
	}

	/* Open ROM file for binary reading */
	fin = fopen(argv[1], "rb");

	/* Self-explanatory... */
	if (fin == NULL) {
		printf("Unable to open input ROM file\n");
		exit(1);
	}

	/* Get current filesize */
	fseek(fin, 0, SEEK_END);
	size = ftell(fin);

	/* Check file size */
	if (size <= ceilp2(HEADER_SIZE)) {
		printf("This file isn't big enough to have been padded!\n");
		exit(1);
	}

	/* Calculate new filesize from original size */
	new_size = ceilp2(size - 1) >> 1;

	/* Move to the beginning of the padding (if there is any) */
	fseek(fin, new_size - HEADER_SIZE, SEEK_SET);

	/* Obtain a sample of the alleged padding for analysis */
	pad1 = getc(fin);

	/* Compare each byte in the input file to the padding byte */
	do {
		i++;
		pad2 = getc(fin);
	} while ((pad1 == pad2) && (pad2 != EOF));

	/* If we got all the way through, it really is padding, and can be removed! */
	if (i != new_size) {
		printf("The ROM is already at its minimum size.\n");
		exit(0);
	}

	if (argc > 2) {
		/* If an output file is specified, open it for binary writing... */
		fout = fopen(argv[3], "wb");
	} else {
		/* ...otherwise, open "temptemp.tmp" for binary writing. */
		fout = fopen(t, "wb");
	}

	/* Self-explanatory... */
	if (fout == NULL) {
		printf("The destination file could not be opened.\n");
		exit(1);
	}

	printf("Attempting to rip %d bytes from your unsuspecting ROM-image...\n", new_size);

	/* Move back to the beginning of the ROM */
	fseek(fin, 0, SEEK_SET);

	/* Copy data from the infile to the outfile */
	for (i = 0; i < new_size - HEADER_SIZE; i++) {
		putc(getc(fin), fout);
	}

	/* Seek to the beginning of the header */
	fseek(fin, -HEADER_SIZE, SEEK_END);

	/* Write the header from the infile. */
	for (i = HEADER_SIZE; i > 0; i--) {
		putc(getc(fin), fout);
	}

	/* We're done!  Shut 'em down! */
	fclose(fin);
	fclose(fout);

	/* If no output file, delete the original, and rename the temp file. */
	if (argc < 3) {
		remove(argv[1]);
		rename(t, argv[1]);
	}

	exit(0);
}
